from .loss import Loss, Reduction
from .categorical_crossentropy import CategoricalCrossentropy

from .mean_squared_error import MeanSquaredError, mean_squared_error
from .sparse_categorical_crossentropy import (
    SparseCategoricalCrossentropy,
    sparse_categorical_crossentropy,
)

from .mean_absolute_error import MeanAbsoluteError, mean_absolute_error
